import json
import logging
import os
import re
import subprocess
import threading
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Tuple

from django.conf import settings
from django.db import transaction
from django.utils import timezone

from .models import Config, Server, SerialHistory, UpdateStatus, WiFiStatus
from .scheme import HostapdConfig


logger = logging.getLogger(__name__)


class KelVPNCliService:
    def __init__(self) -> None:
        self.cli_path: str = "/opt/kelvpn/bin/KelVPNCli"
        self.config_path: str = "/opt/kelvpn/etc/KelVPNCheck.conf"

    def execute_command(
        self, args: List[str], timeout: int = 90, retry_on_timeout: bool = True
    ) -> str:
        """
        Executes a command with the provided arguments and returns the result.
        If a timeout error occurs, it restarts KelVPNService and retries once.

        :param args: List of command arguments.
        :param timeout: Timeout for the command execution in seconds.
        :param retry_on_timeout: Whether to retry the command execution after a timeout.
        :return: Standard output from the command.
        :raises Exception: If the command execution fails after retrying.
        """
        try:
            result = subprocess.run(
                args, capture_output=True, text=True, check=True, timeout=timeout
            )
            output = result.stdout.strip()

            if "The command execution time" in output and retry_on_timeout:
                logger.error(f"Detected timeout error in command output: {output}")
                return self.handle_timeout_and_retry(args, timeout)

            return output

        except subprocess.CalledProcessError as e:
            logger.error(f"Command execution error: {e.stderr.strip()}")
            raise Exception(f"Command execution error: {e.stderr.strip()}") from e
        except subprocess.TimeoutExpired:
            logger.error("Command execution time exceeded the limit.")
            if retry_on_timeout:
                return self.handle_timeout_and_retry(args, timeout)
            raise Exception("Command execution time exceeded the limit.") from None

    def handle_timeout_and_retry(self, args: List[str], timeout: int) -> str:
        """
        Handles command execution timeout by restarting KelVPNService and retrying the command.

        :param args: List of command arguments.
        :param timeout: Timeout for the command execution in seconds.
        :return: Standard output from the retried command.
        :raises Exception: If the retry also fails.
        """
        logger.warning("Restarting KelVPNService due to timeout...")

        cmd_restart_kelvpn: List[str] = [
            "systemctl",
            "restart",
            "KelVPNService.service",
        ]
        logger.debug(
            f"Restarting KelVPNService with command: {' '.join(cmd_restart_kelvpn)}"
        )

        try:
            kelvpn_output: str = self.execute_command(
                cmd_restart_kelvpn, retry_on_timeout=False
            )
            logger.info(f"KelVPNService restarted successfully: {kelvpn_output}")
        except Exception as e:
            logger.error(f"Error during KelVPNService restart: {e}")
            raise Exception(f"Failed to restart KelVPNService: {e}")

        import time

        logger.debug("Sleeping for 2 seconds after restart...")
        time.sleep(2)

        logger.debug(f"Retrying command: {' '.join(args)}")
        try:
            result = subprocess.run(
                args, capture_output=True, text=True, check=True, timeout=timeout
            )
            output = result.stdout.strip()
            logger.info(f"Retry successful. Command output: {output}")
            return output
        except Exception as e:
            logger.error(f"Retry command execution error: {e}")
            raise Exception(f"Retry command execution error: {e}")

    def start_service(self, server_id: str, server_name: str) -> Tuple[str, bool]:
        """
        Starts KelVPNCli with the selected server.

        :param server_id: The ID of the server to connect to.
        :param server_name: The name of the server.
        :return: Tuple containing the command output and a success flag.
        """
        cmd_connect: List[str] = [self.cli_path, "-C", server_id]
        outputs = []

        logger.debug(
            f"Start service called with server_id: {server_id}, server_name: {server_name}"
        )

        # Stop KelVPNCheck.service
        try:
            if self.is_service_active("KelVPNCheck.service"):
                cmd_stop_check: List[str] = ["systemctl", "stop", "KelVPNCheck.service"]
                logger.debug(
                    f"Stopping KelVPNCheck.service with command: {' '.join(cmd_stop_check)}"
                )
                check_service_output: str = self.execute_command(cmd_stop_check)
                outputs.append(f"KelVPNCheck.service stopped: {check_service_output}")
                logger.debug(outputs[-1])
            else:
                logger.debug("KelVPNCheck.service is not active; no need to stop it.")
        except Exception as e:
            if "not loaded" in str(e).lower() or "not found" in str(e).lower():
                outputs.append("KelVPNCheck.service is not running or not found.")
                logger.warning(outputs[-1])
            else:
                outputs.append(f"KelVPNCheck.service stop failed: {e}")
                logger.error(outputs[-1])

        # Attempt connection
        try:
            logger.debug(
                f"Attempting to connect using command: {' '.join(cmd_connect)}"
            )
            output: str = self.execute_command(cmd_connect)
            logger.debug(f"Output from connection attempt: {output}")
        except Exception as e:
            logger.error(f"Command execution error during connection attempt: {e}")
            return f"Command execution error: {e}", False

        # Case: Already connected
        if "There is already a connection to the server." in output:
            logger.info(
                "Detected: Already connected to the server. Updating configuration accordingly."
            )
            self.update_config(
                is_connected=True, is_activated=True, server_name=server_name
            )
            try:
                cmd_start_check: List[str] = [
                    "systemctl",
                    "start",
                    "KelVPNCheck.service",
                ]
                logger.debug(
                    f"Starting KelVPNCheck.service with command: {' '.join(cmd_start_check)}"
                )
                check_service_output: str = self.execute_command(cmd_start_check)
                logger.info(
                    f"KelVPNCheck.service started successfully: {check_service_output}"
                )
            except Exception as e:
                logger.error(
                    f"Error starting KelVPNCheck.service after already connected: {e}"
                )
            return output, True

        # Case: Serial key not found in database
        if "Serial key not found in database" in output:
            logger.warning("Serial key not found in database. Deactivating key.")
            self.deactivate_serial()
            return "Serial key not found and has been deactivated.", False

        # Case: General connection failure (output doesn't contain expected markers)
        if "State: Connect" not in output or "Success!" not in output:
            self.update_config(is_connected=False, server_name=server_name)
            logger.error(f"Connection attempt failed. Output: {output}")
            return output, False

        # Start KelVPNCheck.service
        try:
            cmd_start_check: List[str] = ["systemctl", "start", "KelVPNCheck.service"]
            logger.debug(
                f"Starting KelVPNCheck.service with command: {' '.join(cmd_start_check)}"
            )
            check_service_output: str = self.execute_command(cmd_start_check)
            logger.info(
                f"KelVPNCheck.service started successfully: {check_service_output}"
            )
        except Exception as e:
            logger.error(f"Error starting KelVPNCheck.service: {e}")

        # Successful connection
        self.update_config(
            is_connected=True, is_activated=True, server_name=server_name
        )
        logger.info("Connection established successfully. Updating configuration.")

        return output, True

    def disconnect_service(self) -> str:
        """
        Disconnects KelVPNCli and stops related services.

        :return: Combined output from the commands as a string.
        """
        outputs = []
        try:
            # Disconnect KelVPNCli
            cmd_disconnect: List[str] = [self.cli_path, "-D"]
            disconnect_output: str = self.execute_command(cmd_disconnect)
            outputs.append(f"KelVPNCli disconnected: {disconnect_output}")
            logger.info(outputs[-1])
        except Exception as e:
            outputs.append(f"KelVPNCli disconnect failed: {e}")
            logger.error(outputs[-1])

        # Stop KelVPNCheck.service
        try:
            if self.is_service_active("KelVPNCheck.service"):
                cmd_stop_check: List[str] = ["systemctl", "stop", "KelVPNCheck.service"]
                check_service_output: str = self.execute_command(cmd_stop_check)
                outputs.append(f"KelVPNCheck.service stopped: {check_service_output}")
                logger.info(outputs[-1])
            else:
                outputs.append("KelVPNCheck.service is not running or not found.")
                logger.warning(outputs[-1])
        except Exception as e:
            if "not loaded" in str(e).lower() or "not found" in str(e).lower():
                outputs.append("KelVPNCheck.service is not running or not found.")
                logger.warning(outputs[-1])
            else:
                outputs.append(f"KelVPNCheck.service stop failed: {e}")
                logger.error(outputs[-1])

        # # Stop KelVPNService.service
        # try:
        #     if self.is_service_active("KelVPNService.service"):
        #         cmd_stop_service: List[str] = [
        #             "systemctl",
        #             "stop",
        #             "KelVPNService.service",
        #         ]
        #         service_output: str = self.execute_command(cmd_stop_service)
        #         outputs.append(f"KelVPNService.service stopped: {service_output}")
        #         logger.info(outputs[-1])
        #     else:
        #         outputs.append("KelVPNService.service is not running or not found.")
        #         logger.warning(outputs[-1])
        # except Exception as e:
        #     if "not loaded" in str(e).lower() or "not found" in str(e).lower():
        #         outputs.append("KelVPNService.service is not running or not found.")
        #         logger.warning(outputs[-1])
        #     else:
        #         outputs.append(f"KelVPNService.service stop failed: {e}")
        #         logger.error(outputs[-1])

        # Update configuration
        try:
            self.update_config(is_connected=False)
        except Exception as e:
            logger.error(f"Error updating configuration: {e}")

        return "\n".join(outputs)

    def is_service_active(self, service_name: str) -> bool:
        """
        Checks whether the service exists and whether it is active.
        :param service_name: The name of the service.
        :return: True if the service is active, otherwise False.
        """
        try:
            cmd_check: List[str] = [
                "systemctl",
                "list-units",
                "--full",
                "--all",
                service_name,
            ]
            result: str = self.execute_command(cmd_check).strip()
        except Exception as e:
            logger.warning(f"Service check failed for {service_name}: {e}")
            return False

        if service_name in result:
            try:
                active_result = self.execute_command(
                    ["systemctl", "is-active", service_name]
                ).strip()
                return active_result == "active"
            except Exception as e:
                logger.warning(
                    f"Service is-active check failed for {service_name}: {e}"
                )
                return False
        else:
            logger.warning(f"Service {service_name} is not found in systemd.")
            return False

    def restart_service(self) -> Tuple[str, bool]:
        """
        Restarts the KelVPNService and kelvpn_backend.service.

        :return: Tuple containing the output from the restart commands and success flag.
        """
        try:
            # Update configuration
            self.update_config(is_connected=False)

            # Restart KelVPNService
            cmd_restart_kelvpn: List[str] = [
                "systemctl",
                "restart",
                "KelVPNService.service",
            ]
            kelvpn_output: str = self.execute_command(cmd_restart_kelvpn)
            logger.info(f"KelVPNService restarted successfully: {kelvpn_output}")

            # Restart kelvpn_backend.service
            cmd_restart_backend: List[str] = [
                "systemctl",
                "restart",
                "kelvpn_backend.service",
            ]
            backend_output: str = self.execute_command(cmd_restart_backend)
            logger.info(
                f"kelvpn_backend.service restarted successfully: {backend_output}"
            )

            return kelvpn_output + "\n" + backend_output, True
        except Exception as e:
            logger.error(f"Error restarting services: {str(e)}")
            return str(e), False

    def reboot_system(self) -> str:
        """
        Reboots the system

        :return: Standard output from the reboot command.
        """
        self.update_config(is_connected=False)
        cmd_reboot: List[str] = ["reboot"]
        output: str = self.execute_command(cmd_reboot)
        logger.info("System reboot initiated.")
        return output

    def get_service_status(self) -> str:
        """
        Retrieves the status of the KelVPNService.

        :return: Standard output from the status command.
        """
        cmd_status: List[str] = [
            "systemctl",
            "status",
            "KelVPNService",
            "--no-pager",
        ]
        output: str = self.execute_command(cmd_status)
        return output

    def set_serial(self, serial_key: str) -> Tuple[str, bool, str]:
        """
        Sets the serial key using the best server and updates the remaining days dynamically.

        :param serial_key: The serial key to activate.
        :return: Tuple containing the command output, success flag, and a message indicating the result.
        """
        servers_ping: List[Dict[str, Any]] = self.get_server_pings()
        best_server = min(servers_ping, key=lambda server: server["ping"])
        logger.info(f"best_server {best_server} for activation.")

        if not best_server:
            raise Exception("No server available for activation.")

        server_id = best_server["server_id"]

        logger.info(f"Using best server with ID {server_id} for activation.")

        cmd_set_serial: List[str] = [self.cli_path, "-C", server_id, "-S", serial_key]
        logger.debug(f"Command set serial: {cmd_set_serial}")

        output: str = self.execute_command(cmd_set_serial)
        logger.debug(f"Set serial command output: {output}")

        success, message = self._parse_set_serial_output(output)
        if not success:
            return output, False, message

        remaining_days = self._get_serial_days_left()
        if remaining_days <= 0:
            logger.error("Failed to retrieve remaining days from serial activation.")
            return output, False, "Unable to activate serial key."

        with transaction.atomic():
            config: Optional[Config] = Config.objects.first()

            if not config:
                raise Exception("Configuration not found, cannot proceed.")

            selected_server = Server.objects.filter(server_id=server_id).first()

            if not selected_server:
                raise Exception(
                    f"Selected server with ID {server_id} not found in database."
                )

            config.selected_server = selected_server
            config.is_activated = True
            config.is_connected = True
            config.save()

            logger.info(
                f"Config updated: is_activated=True, is_connected=True, selected_server={selected_server.name}"
            )

            serial, created = SerialHistory.objects.update_or_create(
                serial_key=serial_key,
                defaults={
                    "status": "active",
                    "total_days": remaining_days,
                    "days_active": 0,
                },
            )

            if created:
                logger.info(
                    f"New serial key {serial.serial_key} created with {remaining_days} days."
                )
            else:
                logger.info(
                    f"Existing serial key {serial.serial_key} updated with {remaining_days} days."
                )
                serial.remaining_days = remaining_days
                if serial.status != "active":
                    serial.status = "active"
                    serial.suspended_at = None
                serial.save()

        return output, True, message

    def _get_serial_days_left(self) -> int:
        """
        Retrieves the number of remaining days for the current serial key using KelVPNCli -AI.

        :return: Number of days left or -1 if parsing failed.
        """
        cmd_auth_info: List[str] = [self.cli_path, "-AI"]
        try:
            output = self.execute_command(cmd_auth_info)
            logger.debug(f"Authorization info output: {output}")

            match = re.search(r"days left: (\d+) days", output)
            if match:
                days_left = int(match.group(1))
                logger.info(f"Days left from serial key: {days_left}")
                return days_left

            logger.error("Failed to parse days left from authorization info.")
            return -1

        except Exception as e:
            logger.error(f"Error retrieving days left from serial key: {str(e)}")
            return -1

    def _parse_set_serial_output(self, output: str) -> Tuple[bool, str]:
        """
        Parses the output of the set_serial command to determine success and returns a message.

        :param output: The standard output from the set_serial command.
        :return: Tuple of boolean success flag and message.
        """
        logger.debug(f"Parsing command output (no side effects): {output}")

        if "State: Connect" in output and "Success!" in output:
            return True, "Serial key activated and connected successfully."

        error_match = re.search(r"Error: code: (-?\d+), message: (.+)", output)
        if error_match:
            error_code = error_match.group(1)
            error_msg = error_match.group(2)
            return False, f"Activation failed with error code {error_code}: {error_msg}"

        if "No activated key found" in output and "State: Connecting" in output:
            if "Error: code:" in output:
                error_code = re.search(r"Error: code: (\d+)", output)
                err_msg = f"Activation failed with error code {error_code.group(1) if error_code else 'unknown'}."
                logger.error(err_msg)
                return False, err_msg

        if (
            "Serial key not found in database" in output
            or "Error: code: -31000" in output
        ):
            return False, "Invalid serial key: not found in database."

        if "State: Disconnect" in output and "No activated key found" in output:
            logger.error("No activated key found. Connection attempt failed.")
            return False, "No activated key found. Connection attempt failed."

        if "Invalid serial key format" in output:
            logger.error("Invalid serial key format detected.")
            return False, "Invalid serial key format."

        if "Error: code:" in output:
            error_code = re.search(r"Error: code: (\d+)", output)
            message = re.search(r"message: (.+)", output)
            err_msg = (
                f"Error code {error_code.group(1)}, message: {message.group(1)}"
                if error_code and message
                else "Unknown error occurred."
            )
            logger.error(f"Activation failed: {err_msg}")
            return False, err_msg

        logger.error(f"Unexpected output: {output}")
        return False, "Unexpected error occurred. Check logs for details."

    def _format_serial_key(self, serial: str) -> str:
        """
        Formats a serial key into the database format (XXXX-XXXX-XXXX-XXXX).
        If the serial key already contains hyphens or its length is not 16, returns it unchanged.
        """
        if "-" in serial or len(serial) != 16:
            return serial
        return f"{serial[0:4]}-{serial[4:8]}-{serial[8:12]}-{serial[12:16]}"

    def get_serial_history(self) -> List[Dict[str, Any]]:
        """
        Retrieves and stores the serial history.
        Additionally, checks the authorization info (-AI) and, if the days left for the serial key are negative,
        marks that serial as expired and updates the configuration (is_activated=False).

        :return: List of dictionaries containing serial keys and their creation timestamps.
        """
        cmd_history: List[str] = [self.cli_path, "-H"]
        output: str = self.execute_command(cmd_history)
        serial_keys: List[str] = self._parse_serial_history(output)

        with transaction.atomic():
            for key in serial_keys:
                try:
                    obj, created = SerialHistory.objects.get_or_create(
                        serial_key=key, defaults={"status": "deactivated"}
                    )
                    if created:
                        logger.info(f"Created new SerialHistory for serial key {key}.")
                    else:
                        logger.info(f"Serial key {key} already exists.")
                except Exception as e:
                    logger.error(
                        f"Error while creating or getting SerialHistory for serial key {key}: {str(e)}"
                    )

        # Retrieve authorization info using the -AI command.
        cmd_auth_info: List[str] = [self.cli_path, "-AI"]
        try:
            auth_output: str = self.execute_command(cmd_auth_info)
        except Exception as e:
            logger.error(f"Error executing -AI command: {e}")
            auth_output = ""

        logger.debug(f"Authorization info output: {auth_output}")

        serial_key_match = re.search(
            r"serial key:\s*([A-Z0-9]+)", auth_output, re.IGNORECASE
        )
        days_left_match = re.search(
            r"days left:\s*(Unlimited|[-]?\d+)", auth_output, re.IGNORECASE
        )

        if serial_key_match and days_left_match:
            auth_serial = serial_key_match.group(1).strip()
            auth_serial = self._format_serial_key(auth_serial)
            days_raw = days_left_match.group(1).strip()

            try:
                is_unlimited = days_raw.lower() == "unlimited"
            except AttributeError:
                is_unlimited = False

            if is_unlimited:
                logger.warning(
                    f"Serial key {auth_serial} returned 'Unlimited' — treating as invalid."
                )
                days_left = -1
            else:
                try:
                    days_left = int(days_raw)
                except ValueError:
                    days_left = -1

            logger.info(
                f"Authorization info: serial key = {auth_serial}, days left = {days_left}"
            )

            if days_left < 0:

                try:
                    output, success, deactivated_keys = self.deactivate_serial()
                    if success:
                        logger.info(
                            f"Deactivation command -W executed successfully for {auth_serial}"
                        )
                    else:
                        logger.warning(
                            f"Deactivation command -W failed for {auth_serial}"
                        )
                except Exception as e:
                    logger.error(f"Error during CLI deactivation (-W): {e}")

                try:
                    expired_serial = SerialHistory.objects.filter(
                        serial_key=auth_serial
                    ).first()
                    if expired_serial:
                        expired_serial.status = "expired"
                        expired_serial.save()
                        logger.warning(
                            f"Serial key {auth_serial} is expired or invalid. Days left: {days_left}"
                        )
                    else:
                        logger.warning(
                            f"Serial key {auth_serial} not found in SerialHistory."
                        )
                except Exception as e:
                    logger.error(
                        f"Error updating expiration status for serial key {auth_serial}: {e}"
                    )

                try:
                    config = Config.objects.first()
                    if config:
                        config.is_activated = False
                        config.save()
                        logger.warning(
                            "Config updated: is_activated=False due to expired or invalid serial key."
                        )
                except Exception as e:
                    logger.error(f"Error updating config for expired serial key: {e}")
            else:
                try:
                    serial_obj = SerialHistory.objects.filter(
                        serial_key=auth_serial
                    ).first()
                    if serial_obj:
                        serial_obj.remaining_days = days_left
                        serial_obj.days_active = (
                            serial_obj.total_days - serial_obj.remaining_days
                        )
                        serial_obj.expired_date = serial_obj.created_at + timedelta(
                            days=serial_obj.total_days
                        )
                        serial_obj.save()
                        logger.info(
                            f"Updated remaining days for serial key {auth_serial} to {days_left}."
                        )
                    else:
                        logger.warning(
                            f"Serial key {auth_serial} not found in SerialHistory."
                        )
                except Exception as e:
                    logger.error(
                        f"Error updating remaining days for serial key {auth_serial}: {e}"
                    )

        history: List[SerialHistory] = SerialHistory.objects.all().order_by(
            "-created_at"
        )
        logger.info(f"Serial history: {list(history)}")
        return [
            {
                "serial_key": entry.serial_key,
                "created_at": entry.created_at,
                "status": entry.status,
                "total_days": entry.total_days,
                "days_active": entry.days_active,
                "remaining_days": entry.remaining_days,
                "expired_date": entry.expired_date,
            }
            for entry in history
        ]

    def _parse_serial_history(self, output: str) -> List[str]:
        """
        Parses the output of the KelVPN -H command and returns a list of serial keys.

        :param output: The standard output from the KelVPNCli -H command.
        :return: List of serial keys.
        """
        serial_keys: List[str] = []
        lines: List[str] = output.splitlines()

        # Find the line containing "Serial History size"
        size_line: Optional[str] = next(
            (line for line in lines if "Serial History size" in line), None
        )
        if size_line:
            try:
                size: int = int(
                    re.search(r"Serial History size: (\d+)", size_line).group(1)
                )
            except (AttributeError, ValueError):
                size = 0
        else:
            size = 0

        for line in lines:
            match = re.match(r"\d{2} : ([A-Z0-9-]{19})", line)
            if match:
                serial_keys.append(match.group(1))

        return serial_keys

    def deactivate_serial(self) -> Tuple[str, bool, Optional[List[str]]]:
        """
        Deactivates all active serial key(s) by resetting them using KelVPNCli (-W) and updates the serial history.

        If the output indicates success (or that there are no activated keys),
        all active serial keys are marked as deactivated and the configuration is updated (is_activated=False).

        :return: Tuple containing the command output, success flag, and a list of deactivated serial key strings (or None if none).
        """
        cmd_reset_serial: List[str] = [self.cli_path, "-W"]
        output: str = self.execute_command(cmd_reset_serial)
        logger.debug(f"Deactivation command output: {output}")

        if ("Success!" in output and "Resetting the serial key." in output) or (
            "There are no activated keys." in output
        ):
            active_serials = list(SerialHistory.objects.filter(status="active"))
            deactivated_keys = []
            if active_serials:
                for serial in active_serials:
                    serial.status = "deactivated"
                    serial.suspended_at = timezone.now()
                    serial.save()
                    deactivated_keys.append(serial.serial_key)
                    logger.info(
                        f"Serial key {serial.serial_key} deactivated successfully."
                    )
            else:
                logger.warning("No active serial keys found during deactivation.")

            self.update_config(is_connected=False, is_activated=False)
            return output, True, deactivated_keys if deactivated_keys else None

        logger.error(f"Failed to deactivate serial key. Command output: {output}")
        return output, False, None

    def get_server_pings(self) -> List[Dict[str, Any]]:
        servers_output: str = self.execute_command([self.cli_path, "-L"])
        actual_servers = {}
        server_line_regex = re.compile(r"\s*(\d{2})\s*\|\s*(.+?)\s*$")
        for line in servers_output.splitlines():
            match = server_line_regex.match(line)
            if match:
                server_id, server_name = match.group(1).strip(), match.group(2).strip()
                actual_servers[server_name] = server_id

        ping_output: str = self.execute_command([self.cli_path, "-L", "-R"])
        ping_line_regex = re.compile(r"\s*\d{2}\s*\|\s*(\d+)\s*\|\s*(.+?)\s*$")

        servers_ping = []
        for line in ping_output.splitlines():
            match = ping_line_regex.match(line)
            if match:
                ping = int(match.group(1).strip())
                server_name = match.group(2).strip()

                actual_server_id = actual_servers.get(server_name)
                if actual_server_id:
                    servers_ping.append(
                        {
                            "server_id": actual_server_id,
                            "name": server_name,
                            "ping": ping,
                            "selected": False,
                        }
                    )

        with transaction.atomic():
            config = Config.objects.select_related("selected_server").first()
            existing_selected_server_id = (
                config.selected_server.server_id
                if config and config.selected_server
                else None
            )

            for server in servers_ping:
                is_selected = server["server_id"] == existing_selected_server_id

                server_obj, _ = Server.objects.update_or_create(
                    server_id=server["server_id"],
                    defaults={
                        "name": server["name"],
                        "ping": server["ping"],
                        "selected": is_selected,
                    },
                )

                if is_selected:
                    server["selected"] = True

            if (
                not config.selected_server
                or not Server.objects.filter(
                    server_id=existing_selected_server_id
                ).exists()
            ):
                Server.objects.filter(selected=True).update(selected=False)
                best_server = min(servers_ping, key=lambda s: s["ping"], default=None)
                if best_server:
                    selected_server = Server.objects.get(
                        server_id=best_server["server_id"]
                    )
                    selected_server.selected = True
                    selected_server.save()
                    config.selected_server = selected_server
                    config.save()

        return sorted(servers_ping, key=lambda s: s["ping"])

    def _parse_server_ping_list(self, output: str) -> List[Dict[str, Any]]:
        """
        Parses the output of the KelVPNCli -L -R command and returns a list of servers with pings.

        :param output: The standard output from the KelVPNCli -L -R command.
        :return: List of dictionaries containing server details.
        """
        servers_ping: List[Dict[str, Any]] = []
        lines: List[str] = output.splitlines()

        server_ping_line_regex = re.compile(
            r"^\s*(\d{2})\s*\|\s*(\d{1,4})\s*\|\s*(.+?)\s*$"
        )

        for line in lines:
            match = server_ping_line_regex.match(line)
            if match:
                server_id: str = match.group(1).strip()
                ping_str: str = match.group(2).strip()
                server_name: str = match.group(3).strip()

                try:
                    ping: int = int(ping_str)
                except ValueError:
                    ping = -1

                if server_id and server_name:
                    servers_ping.append(
                        {
                            "id": server_id,
                            "ping": ping,
                            "name": server_name,
                            "selected": False,
                        }
                    )
                    logger.debug(
                        f"Parsed server with ping: ID={server_id}, Ping={ping}, Name={server_name}"
                    )

        logger.info(f"Total servers with ping parsed: {len(servers_ping)}")
        return servers_ping

    def get_location(self) -> str:
        try:
            response = subprocess.check_output(
                ["curl", "-s", "http://ip-api.com/json/"]
            )
            json_data = json.loads(response.decode("utf-8"))
            country = json_data.get("country", "-")
            city = json_data.get("city", "")
            return f"{country}, {city}" if city else country
        except Exception as e:
            logger.warning(f"Failed to fetch location info: {e}")
            return "-"

    def update_config(
        self,
        is_connected: bool,
        is_activated: Optional[bool] = None,
        server_name: Optional[str] = None,
    ) -> None:
        """
        Updates the config's connection and activation status.

        :param is_connected: Boolean indicating if the service is connected.
        :param is_activated: Optional boolean indicating if the service is activated.
        """
        config: Optional[Config] = Config.objects.first()
        location = self.get_location()

        if config:
            config.is_connected = is_connected
            config.location = location
            if is_activated is not None:
                config.is_activated = is_activated
            config.save()

        try:
            with open(self.config_path, "r") as file:
                lines = file.readlines()

            new_lines = []
            found_connected = False
            found_server = False

            for line in lines:
                if line.startswith("KELVPN_CONNECTED="):
                    new_lines.append(
                        f'KELVPN_CONNECTED={"true" if is_connected else "false"}\n'
                    )
                    found_connected = True
                elif line.startswith("KELVPN_SERVER=") and server_name:
                    new_lines.append(f'KELVPN_SERVER="{server_name}"\n')
                    found_server = True
                else:
                    new_lines.append(line)

            if not found_connected:
                new_lines.append(
                    f'KELVPN_CONNECTED={"true" if is_connected else "false"}\n'
                )
            if not found_server and server_name:
                new_lines.append(f'KELVPN_SERVER="{server_name}"\n')

            with open(self.config_path, "w") as file:
                file.writelines(new_lines)

            logger.info(
                f"Updated configuration: KELVPN_CONNECTED={'true' if is_connected else 'false'}, KELVPN_SERVER={server_name}"
            )

        except Exception as e:
            logger.error(f"Failed to update configuration: {e}")

    def _get_ip_addresses(self):
        # result = subprocess.run(["nmcli", "device", "show"], capture_output=True, text=True)
        # addresses = re.findall(r"IP4\.ADDRESS\[\d+\]:\s+([\d.]+)", result.stdout)
        try:
            external_ip = (
                subprocess.check_output(["curl", "-s", "ifconfig.me"]).decode().strip()
            )
            # logger.debug(f"ip {external_ip}")
            return external_ip

        except subprocess.CalledProcessError as e:
            logger.debug(f"Error retrieving external IP: {e}")
            return None

    def _get_gateway(self):
        """
        Получает IP-адрес шлюза с приоритетом на tun0 (VPN), затем eth0 (Ethernet) и, наконец, wlan0 (Wi-Fi).
        """
        result = subprocess.run(
            ["nmcli", "device", "show"], capture_output=True, text=True
        )
        output = result.stdout

        # Определяем приоритет интерфейсов
        priority_interfaces = ["tun0", "eth0", "wlan0"]

        for interface in priority_interfaces:
            if f"GENERAL.DEVICE:                         {interface}" in output:
                match = re.search(
                    rf"{interface}\n(?:.*\n)*?IP4\.GATEWAY:\s+([\d.]+)", output
                )
                if match:
                    return match.group(1)

        return "-"

    def _get_dns_servers(self):
        dns_servers = []
        try:
            with open("/etc/resolv.conf", "r") as file:
                for line in file:
                    if line.startswith("nameserver"):
                        dns_servers.append(line.split()[1])
        except Exception as e:
            logger.error(f"Error retrieving DNS servers from resolv.conf: {str(e)}")
        return dns_servers

    def update_network_config(self):
        """
        Updates network configuration and remaining days of VPN usage.
        """
        ip_addresses = self._get_ip_addresses()
        gateway = self._get_gateway()
        dns_servers = self._get_dns_servers()

        logger.info(
            f"Updating network config: ip={ip_addresses}, gateway={gateway}, dns_servers={dns_servers}"
        )

        # Get or create the configuration record
        config, created = Config.objects.get_or_create(id=1)

        config.ip = ip_addresses if ip_addresses else "-"
        config.gateway = gateway
        config.dns = ", ".join(dns_servers)

        # Check when the last active serial was updated
        latest_serial = SerialHistory.objects.filter(status="active").first()

        if latest_serial:
            last_updated = (
                latest_serial.updated_at
                if latest_serial.updated_at
                else latest_serial.created_at
            )
            time_since_last_update = timezone.now() - last_updated

            # Update only if more than 1 hour has passed
            if time_since_last_update >= timedelta(hours=1):
                logger.info(
                    "Fetching latest serial history to update remaining days..."
                )
                self.get_serial_history()

                # Fetch the latest serial again after updating the history
                latest_serial = (
                    SerialHistory.objects.filter(status="active")
                    .order_by("-created_at")
                    .first()
                )
                if latest_serial:
                    config.remaining_days = latest_serial.remaining_days
                    logger.info(
                        f"Updated remaining_days in Config: {config.remaining_days}"
                    )
                else:
                    logger.warning(
                        "No serial history found. Keeping previous remaining_days."
                    )
            else:
                logger.info(
                    f"Skipping serial history update. Last updated {time_since_last_update.seconds // 60} minutes ago."
                )
        else:
            logger.warning("No active serial found. Skipping serial update.")

        config.save()
        logger.info("Network configuration updated successfully.")


class WifiManager:
    CONFIG_PATHS = settings.HOSTAPD_CONFIG_PATHS

    @staticmethod
    def start_hostapd(interface: str) -> None:
        """
        Starts the hostapd service for the given interface.
        """
        subprocess.run(["systemctl", "start", f"hostapd@{interface}"], check=True)

    @staticmethod
    def stop_hostapd(interface: str) -> None:
        """
        Stops the hostapd service for the given interface.
        """
        subprocess.run(["systemctl", "stop", f"hostapd@{interface}"], check=True)

    @staticmethod
    def restart_hostapd(interface: str) -> None:
        """
        Restarts the hostapd service for the given interface.
        """
        subprocess.run(["systemctl", "restart", f"hostapd@{interface}"], check=True)

    @staticmethod
    def toggle_hostapd(interface: str, enable: bool) -> None:
        """
        Toggles the hostapd service based on the `enable` flag.
        """
        if enable:
            WifiManager.start_hostapd(interface)
        else:
            WifiManager.stop_hostapd(interface)

    @staticmethod
    def read_configs() -> list[dict]:
        """
        Reads configurations from all hostapd config files.
        Returns a list of configuration dictionaries.
        """
        configs = []
        for interface, config_path in WifiManager.CONFIG_PATHS.items():
            config = {"interface": interface, "config_file": config_path}
            try:
                with open(config_path, "r") as file:
                    for line in file:
                        line = line.strip()
                        if line and not line.startswith("#") and "=" in line:
                            key, value = line.split("=", 1)
                            config[key.strip()] = value.strip()
                configs.append(config)
            except FileNotFoundError:
                logger.error(f"Configuration file {config_path} not found.")
        return configs

    @staticmethod
    def determine_band(config: dict) -> str:
        """
        Determines the Wi-Fi band based on the configuration.
        """
        hw_mode = config.get("hw_mode", "g")
        channel = int(config.get("channel", 0))

        if hw_mode == "g" and 1 <= channel <= 14:
            return "2.4GHz"
        if hw_mode == "a" and 36 <= channel <= 165:
            return "5GHz"
        if hw_mode == "6" and 1 <= channel <= 233:
            return "6GHz"
        return "Unknown"

    @staticmethod
    def update_wifi_status() -> None:
        """
        Updates the WiFiStatus records based on configurations.
        """
        configs = WifiManager.read_configs()
        for config in configs:
            try:
                WiFiStatus.objects.update_or_create(
                    interface=config["interface"],
                    defaults={
                        "band": WifiManager.determine_band(config),
                        "ssid": config.get("ssid", "-"),
                        "channel": config.get("channel", "-"),
                        "hw_mode": config.get("hw_mode", "-"),
                        "transmit_power": config.get("transmit_power", 0),
                        "enabled": WifiManager.check_service_status(
                            config["interface"]
                        ),
                        "wpa_passphrase": config.get("wpa_passphrase", "-"),
                    },
                )
            except Exception as e:
                logger.error(f"Failed to update Wi-Fi status: {e}")

    @staticmethod
    def check_service_status(interface: str) -> bool:
        """
        Checks whether the hostapd service for the given interface is active.
        """
        result = subprocess.run(
            ["systemctl", "is-active", f"hostapd@{interface}"],
            stdout=subprocess.PIPE,
            text=True,
        )
        return result.stdout.strip() == "active"

    @staticmethod
    def update_network_config(interface: str, **kwargs) -> None:
        """
        Updates Wi-Fi network configuration parameters in the hostapd config file.
        Only provided kwargs are updated.
        """
        config_path = WifiManager.CONFIG_PATHS.get(interface)
        if not config_path:
            raise FileNotFoundError(f"Configuration file for {interface} not found.")

        config_data = {}
        if os.path.exists(config_path):
            with open(config_path, "r") as file:
                for line in file:
                    line = line.strip()
                    if line and not line.startswith("#") and "=" in line:
                        key, value = line.split("=", 1)
                        config_data[key.strip()] = value.strip()

        for key, value in kwargs.items():
            if value is not None:
                config_data[key] = str(value)

        with open(config_path, "w") as file:
            for key, value in config_data.items():
                file.write(f"{key}={value}\n")

        WifiManager.restart_hostapd(interface)


class UpdateService:
    @staticmethod
    def check_updates():
        def get_deb_version(package: str) -> Optional[str]:
            try:
                result = subprocess.run(
                    ["dpkg", "-s", package],
                    capture_output=True,
                    text=True,
                    check=True,
                )
                for line in result.stdout.splitlines():
                    if line.startswith("Version:"):
                        return line.split(":", 1)[1].strip()
            except subprocess.CalledProcessError:
                return None

        kelvpn_version = get_deb_version("kelvpn")
        backend_version = get_deb_version("kelvpn-admin")

        try:
            result = subprocess.run(
                ["apt-get", "-s", "upgrade"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            updates_available = not (
                "0 upgraded, 0 newly installed" in result.stdout
                or "обновлено 0 пакетов, установлено 0 новых пакетов"
                in result.stdout.lower()
            )
        except Exception as e:
            logger.error(f"Error checking for updates: {str(e)}")
            raise

        return {
            "updates_available": updates_available,
            "kelvpn_version": kelvpn_version or "unknown",
            "backend_version": backend_version or "unknown",
        }

    @staticmethod
    def perform_update():
        update_status, created = UpdateStatus.objects.get_or_create(id=1)
        log_path = "/opt/kelvpn-web-admin/backend/logs/"
        if update_status.status == "IN_PROGRESS":
            return False, "Update already in progress"

        update_status.status = "IN_PROGRESS"
        update_status.log_file = (
            f"{log_path}update_log_{datetime.now().strftime('%Y%m%d_%H%M%S')}.txt"
        )
        update_status.save()

        def update_process():
            try:
                with open(update_status.log_file, "w") as f:
                    subprocess.run(
                        ["sudo", "apt-get", "update", "-y"],
                        stdout=f,
                        stderr=f,
                        text=True,
                    )

                    result_upgrade = subprocess.run(
                        ["sudo", "apt-get", "upgrade", "-y"],
                        stdout=f,
                        stderr=f,
                        text=True,
                    )

                    if result_upgrade.returncode != 0:
                        update_status.status = "FAILED"
                        f.write(f"Error during upgrade: {result_upgrade.stderr}\n")
                    else:
                        update_status.status = "COMPLETED"

                    update_status.completed_at = datetime.now()
                    update_status.save()

            except Exception as e:
                with open(update_status.log_file, "a") as f:
                    f.write(f"Error performing update: {str(e)}\n")
                update_status.status = "FAILED"
                update_status.completed_at = datetime.now()
                update_status.save()

        # Запуск задачи в фоновом режиме
        thread = threading.Thread(target=update_process)
        thread.start()

        return True, "Update started"

    @staticmethod
    def get_update_status():
        update_status, created = UpdateStatus.objects.get_or_create(id=1)
        return {
            "status": update_status.status,
            "log_file": f"logs/{update_status.log_file}",
            "started_at": update_status.started_at,
            "completed_at": update_status.completed_at,
        }


class NetworkManager:
    WAN_INTERFACE = "wan0"

    @staticmethod
    def get_network_settings():
        settings = {
            "mac_address": NetworkManager.get_mac_address(),
            "ip_address": NetworkManager.get_ip_address(),
            "gateway": NetworkManager.get_default_gateway(),
            "ip_mode": NetworkManager.get_ip_mode(),
        }
        return settings

    @staticmethod
    def get_mac_address():
        result = subprocess.run(
            ["cat", f"/sys/class/net/{NetworkManager.WAN_INTERFACE}/address"],
            stdout=subprocess.PIPE,
            text=True,
        )
        return result.stdout.strip()

    @staticmethod
    def set_mac_address(mac_address: str):
        if not re.match(r"^([0-9A-Fa-f]{2}:){5}[0-9A-Fa-f]{2}$", mac_address):
            raise ValueError("Invalid MAC address format.")

        subprocess.run(["ip", "link", "set", NetworkManager.WAN_INTERFACE, "down"])
        subprocess.run(
            ["ip", "link", "set", NetworkManager.WAN_INTERFACE, "address", mac_address]
        )
        subprocess.run(["ip", "link", "set", NetworkManager.WAN_INTERFACE, "up"])

    @staticmethod
    def get_ip_address():
        result = subprocess.run(
            ["ip", "-4", "addr", "show", NetworkManager.WAN_INTERFACE],
            stdout=subprocess.PIPE,
            text=True,
        )
        match = re.search(r"inet (\d+\.\d+\.\d+\.\d+)", result.stdout)
        return match.group(1) if match else "Not Assigned"

    @staticmethod
    def get_default_gateway():
        result = subprocess.run(
            ["ip", "route", "show", "default"],
            stdout=subprocess.PIPE,
            text=True,
        )
        match = re.search(r"default via (\d+\.\d+\.\d+\.\d+)", result.stdout)
        return match.group(1) if match else "Not Assigned"

    @staticmethod
    def get_ip_mode():
        dhcp_file = f"/var/lib/dhcp/dhclient.{NetworkManager.WAN_INTERFACE}.leases"
        if os.path.exists(dhcp_file):
            return "Dynamic"
        return "Static"

    @staticmethod
    def set_ip_config(static: bool, ip_address=None, netmask=None, gateway=None):
        if static:
            if not all([ip_address, netmask, gateway]):
                raise ValueError("IP address, netmask, and gateway must be set.")
            subprocess.run(["ip", "addr", "flush", "dev", NetworkManager.WAN_INTERFACE])
            subprocess.run(
                [
                    "ip",
                    "addr",
                    "add",
                    f"{ip_address}/{netmask}",
                    "dev",
                    NetworkManager.WAN_INTERFACE,
                ]
            )
            subprocess.run(["ip", "route", "add", "default", "via", gateway])
        else:
            subprocess.run(["dhclient", NetworkManager.WAN_INTERFACE])
