import logging
import json
from typing import List

from django.db import transaction
from django.shortcuts import get_object_or_404
from django.contrib.auth import authenticate
from django.contrib.auth.hashers import check_password, make_password
from rest_framework_simplejwt.tokens import RefreshToken
from ninja import Router
from ninja.errors import HttpError

from .scheme import (
    SetSerial,
    ServerSchema,
    ConfigSchema,
    UpdateConfigSchema,
    SelectServerSchema,
    StartServiceResponse,
    DisconnectServiceResponse,
    RestartServiceResponse,
    AuthSchema,
    AuthResponse,
    UpdateUserSchema,
    HostapdConfig,
    WiFiStatusSchema,
    UpdateWiFiStatusSchema,
    WorkProfileSchema,
    CreateProfileSchema,
    ApplicationSchema,
    UpdateResponse,
    UpdateStatusResponse,
    UpdateNetworkSchema,
)

from .models import (
    Server,
    Config,
    WiFiStatus,
    WorkProfile,
    ProfileURL,
    Application,
    SerialHistory,
)
from .auth import TokenAuth
from .services import KelVPNCliService, WifiManager, UpdateService

logger = logging.getLogger(__name__)

router = Router()
vpn_service = KelVPNCliService()
update_service = UpdateService()


@router.post("/auth", response=AuthResponse)
def auth(request, auth_data: AuthSchema):
    user = authenticate(username=auth_data.username, password=auth_data.password)

    if user is None:
        raise HttpError(401, "Error in credentials")

    refresh = RefreshToken.for_user(user)
    access = refresh.access_token

    return {
        "access": str(access),
        "refresh": str(refresh),
    }


@router.put("/user/update", auth=TokenAuth())
def update_user(request, payload: UpdateUserSchema):
    """
    Updates the user's login and password after verifying the current password.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized.")

        if not check_password(payload.current_password, user.password):
            raise HttpError(400, "Current password is incorrect.")

        if payload.username and payload.username != user.username:
            user.username = payload.username

        if payload.password:
            user.password = make_password(payload.password)

        user.save()

        return {"message": "User data updated successfully."}

    except HttpError as e:
        raise e
    except Exception as e:
        logger.error(f"Unexpected error while updating user: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.post("/token/refresh", response=AuthResponse)
def refresh_token(request, refresh_token: str):
    try:
        refresh = RefreshToken(refresh_token)
        access = refresh.access_token
        return {"access": str(access), "refresh": str(refresh)}
    except Exception as e:
        raise HttpError(400, f"Error refresh token: {e}")


@router.post("/service/start", response=StartServiceResponse, auth=TokenAuth())
def start_service(request):
    """
    Launching KelVPN with the selected server from the configuration.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized")
        config = Config.objects.first()
        if not config or not config.selected_server:
            raise HttpError(400, "No server selected in configuration.")

        selected_server_id = config.selected_server.server_id
        selected_server_name = config.selected_server.name
        logger.info(f"selected server: {selected_server_id} - {selected_server_name}")
        output, success = vpn_service.start_service(
            selected_server_id, selected_server_name
        )

        if not success:
            raise HttpError(500, f"Failed to connect to the server. Output: {output}")

        return StartServiceResponse(
            message="KelVPN has been launched successfully.",
            output=output,
            error=None,
        )
    except HttpError as e:
        logger.error(f"HttpError in start_service: {e!r}")
        raise e
    except Exception as e:
        logger.error(f"Unexpected error in start_service: {e!r}")
        raise HttpError(500, "Internal Server Error.")


@router.post(
    "/service/disconnect", response=DisconnectServiceResponse, auth=TokenAuth()
)
def disconnect_service(request):
    """
    Disabling KelVPN.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized")

        output = vpn_service.disconnect_service()

        return DisconnectServiceResponse(
            message="KelVPN has been disabled successfully.",
            output=output,
            error=None,
        )
    except Exception as e:
        logger.error(f"Unexpected error when disabling the service: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.post("/service/restart", response=RestartServiceResponse, auth=TokenAuth())
def restart_service(request):
    """
    Restarting the KelVPNService.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized")

        output = vpn_service.restart_service()

        return RestartServiceResponse(
            message="KelVPNService has been restarted successfully.",
            output=output,
            error=None,
        )
    except Exception as e:
        logger.error(f"Unexpected error during service restart: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.post("/system/reboot", auth=TokenAuth())
def reboot_system(request):
    """
    Reboots the system.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized.")

        output = vpn_service.reboot_system()

        return {"message": "System reboot initiated successfully.", "output": output}

    except Exception as e:
        logger.error(f"Error initiating system reboot: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.get("/service/status", auth=TokenAuth())
def get_service_status(request):
    """
    Retrieves the status of the KelVPNService.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized.")

        status = vpn_service.get_service_status()
        return {"status": status}
    except Exception as e:
        logger.error(f"Error retrieving service status: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.post("/service/set-serial", auth=TokenAuth())
def connect_service(request, payload: SetSerial):
    """
    Launches KelVPN with the specified serial key.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized.")

        serial_key = payload.serial_key.strip()
        output, success, message = vpn_service.set_serial(serial_key)

        if not success:
            logger.error(f"Failed to set serial key: {message}")
            raise HttpError(500, message)

        return StartServiceResponse(
            message=message,
            output=output,
            error=None,
        )

    except HttpError:
        raise
    except Exception as e:
        logger.error(f"Unexpected error in connect_service: {str(e)}")
        raise HttpError(500, "Unexpected internal error. Please try again.")


@router.post(
    "/service/deactivate-serial", response=DisconnectServiceResponse, auth=TokenAuth()
)
def deactivate_serial_service(request):
    """
    Deactivates the serial key and updates the configuration status.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized.")

        output, success, deactivated_serial = vpn_service.deactivate_serial()

        if not success:
            return DisconnectServiceResponse(
                message="The command is executed, but the serial key was not deactivated.",
                output=output,
                error="Failed to deactivate the serial key.",
            )

        if deactivated_serial:
            keys_str = ", ".join(deactivated_serial)
        else:
            keys_str = "None"

        return DisconnectServiceResponse(
            message=f"The serial keys {keys_str} were deactivated successfully.",
            output=output,
            error=None,
            serial_key=keys_str,
            total_days=0,
            remaining_days=0,
            status="deactivated",
        )

    except Exception as e:
        logger.error(f"Unexpected error in deactivate_serial_service: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.get("/service/serial-history", auth=TokenAuth())
def get_serial_history(request):
    """
    Retrieves the serial history.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized.")

        history = vpn_service.get_serial_history()
        return {"serial_history": history}
    except Exception as e:
        logger.error(f"Error retrieving serial history: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.get("/servers-ping", response=List[ServerSchema], auth=TokenAuth())
def get_servers_ping(request):
    """
    Retrieves the list of servers with their pings and selects the best server if none is selected.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized.")

        servers = vpn_service.get_server_pings()
        return servers
    except Exception as e:
        logger.error(f"Error retrieving servers ping: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.get("/config", response=ConfigSchema, auth=TokenAuth())
def get_config(request):
    """
    Retrieves the current configuration, updating it from the system first.
    """
    try:
        user = request.auth
        if not user:
            raise HttpError(401, "Not authorized.")

        kelvpn_service = KelVPNCliService()
        kelvpn_service.update_network_config()

        config = Config.objects.first()
        if not config:
            raise HttpError(404, "Configuration not found.")

        active_serial = SerialHistory.objects.filter(status="active").first()
        remaining_days = (
            active_serial.remaining_days if active_serial else config.remaining_days
        )

        selected_server = config.selected_server
        server_schema = (
            ServerSchema(
                server_id=selected_server.server_id,
                name=selected_server.name,
                ping=selected_server.ping,
                selected=selected_server.selected,
            )
            if selected_server
            else None
        )

        config_schema = ConfigSchema(
            ip=config.ip,
            location=config.location,
            dns=config.dns,
            gateway=config.gateway,
            is_activated=config.is_activated,
            is_connected=config.is_connected,
            remaining_days=remaining_days,
            selected_server=server_schema,
        )
        return config_schema

    except Exception as e:
        logger.error(f"Error retrieving config: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.put("/config", response=ConfigSchema, auth=TokenAuth())
def update_config(request, data: UpdateConfigSchema):
    """
    Updates the configuration.
    """
    try:
        with transaction.atomic():
            config = Config.objects.first()
            if not config:
                config = Config.objects.create(
                    ip=data.ip or "0.0.0.0",
                    location=data.location or "",
                    dns=data.dns or "8.8.8.8",
                    gateway=data.gateway or "192.168.100.1",
                )

            # Update configuration fields
            if data.ip:
                config.ip = data.ip
            if data.location:
                config.location = data.location
            if data.dns:
                config.dns = data.dns
            if data.gateway:
                config.gateway = data.gateway

            # Update selected server if provided
            if data.selected_server_id:
                try:
                    selected_server = Server.objects.get(
                        server_id=data.selected_server_id
                    )
                    # Deselect all other servers
                    Server.objects.filter(selected=True).update(selected=False)
                    # Select the new server
                    selected_server.selected = True
                    selected_server.save()
                    config.selected_server = selected_server
                except Server.DoesNotExist:
                    raise HttpError(
                        400, {"error": "Server with the given ID does not exist."}
                    )

            config.save()

        # Prepare response
        selected_server = config.selected_server
        server_schema = (
            ServerSchema(
                server_id=selected_server.server_id,
                name=selected_server.name,
                ping=selected_server.ping,
                selected=selected_server.selected,
            )
            if selected_server
            else None
        )

        config_schema = ConfigSchema(
            ip=config.ip,
            location=config.location,
            dns=config.dns,
            gateway=config.gateway,
            is_activated=config.is_activated,
            is_connected=config.is_connected,
            selected_server=server_schema,
        )
        return config_schema
    except HttpError as he:
        raise he
    except Exception as e:
        logger.error(f"Error updating config: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.post("/select-server", response=ServerSchema, auth=TokenAuth())
def select_server(request, payload: SelectServerSchema):
    """
    Selects a server based on the provided server ID.
    """
    try:
        server = Server.objects.get(server_id=payload.server_id)

        with transaction.atomic():
            # Deselect all other servers
            Server.objects.filter(selected=True).update(selected=False)

            # Select the desired server
            server.selected = True
            server.save()

            # Update configuration
            config, created = Config.objects.get_or_create(
                id=1,
                defaults={
                    "ip": "0.0.0.0",
                    "location": "Unknown",
                    "dns": "8.8.8.8",
                    "gateway": "192.168.100.1",
                    "selected_server": server,
                },
            )
            if not created:
                config.selected_server = server
                config.save()

        disconnect_output = vpn_service.disconnect_service()
        logger.info(
            f"Service disconnected after changing the server. Output: {disconnect_output}"
        )

        server_data = ServerSchema(
            server_id=server.server_id,
            name=server.name,
            ping=server.ping,
            selected=server.selected,
        )

        return server_data
    except Server.DoesNotExist:
        raise HttpError(404, {"error": "Server not found."})
    except Exception as e:
        logger.error(f"Error selecting server: {str(e)}")
        raise HttpError(500, {"error": "Internal Server Error."})


@router.get("/wifi/status", response=List[WiFiStatusSchema])
def get_wifi_status(request):
    """
    Получает список текущих состояний Wi-Fi и их настройки.
    """
    WifiManager.update_wifi_status()
    statuses = WiFiStatus.objects.all()
    return statuses


@router.put("/wifi/status/{band}", response=WiFiStatusSchema)
def update_wifi_status(request, band: str, payload: UpdateWiFiStatusSchema):
    """
    Обновляет состояние частоты Wi-Fi (включение или отключение).
    """
    wifi_status = get_object_or_404(WiFiStatus, band=band)
    wifi_status.enabled = payload.enabled
    wifi_status.save()

    WifiManager.toggle_hostapd(wifi_status.interface, payload.enabled)
    return wifi_status


@router.get("/wifi/config/{interface}", response=HostapdConfig)
def get_wifi_config(request, interface: str):
    """
    Получает конфигурацию Wi-Fi для указанного интерфейса.
    """
    config = WifiManager.read_config(interface)
    return config


@router.put("/wifi/config", response=HostapdConfig)
def update_wifi_config(request, interface: str, payload: HostapdConfig):
    """
    Обновляет конфигурацию Wi-Fi.
    """
    WifiManager.update_config(interface, payload)
    updated_config = WifiManager.read_config(interface)
    return updated_config


@router.put("/wifi/{network_id}/update")
def update_wifi_network(request, network_id: int, payload: UpdateNetworkSchema):
    """
    Обновляет конфигурацию Wi-Fi по идентификатору сети.
    """
    try:
        wifi_status = WiFiStatus.objects.get(id=network_id)
        WifiManager.update_network_config(
            wifi_status.interface,
            ssid=payload.ssid,
            hw_mode=payload.hw_mode,
            channel=payload.channel,
            wpa_passphrase=payload.wpa_passphrase,
        )

        for field, value in payload.dict(exclude_unset=True).items():
            setattr(wifi_status, field, value)
        wifi_status.save()

        return {"message": "Конфигурация успешно обновлена."}

    except WiFiStatus.DoesNotExist:
        raise HttpError(404, "Сеть Wi-Fi не найдена.")
    except Exception as e:
        logger.error(f"Ошибка обновления конфигурации Wi-Fi: {e}")
        raise HttpError(500, "Ошибка при обновлении конфигурации.")


@router.post("/profiles/create", response=WorkProfileSchema, auth=TokenAuth())
def create_profile(request, payload: CreateProfileSchema):
    """
    Creates a new work profile. If the new profile is set to active, all other profiles will be deactivated.
    """
    try:
        if payload.is_active:
            WorkProfile.objects.filter(is_active=True).update(is_active=False)

        profile = WorkProfile.objects.create(
            name=payload.name,
            profile_type=payload.profile_type,
            is_active=payload.is_active,
        )

        # Обрабатываем URL-адреса: создаем или получаем записи ProfileURL
        for url in payload.selected_urls:
            profile_url, _ = ProfileURL.objects.get_or_create(url=url)
            profile.selected_urls.add(profile_url)

        # Обрабатываем приложения: создаем или получаем записи Application
        for app_name in payload.selected_apps:
            app, _ = Application.objects.get_or_create(name=app_name)
            profile.selected_apps.add(app)

        profile.save()
        return {
            "id": profile.id,
            "name": profile.name,
            "profile_type": profile.profile_type,
            "is_active": profile.is_active,
            "selected_urls": [url.url for url in profile.selected_urls.all()],
            "selected_apps": [app.name for app in profile.selected_apps.all()],
        }
    except Exception as e:
        logger.error(f"Error creating profile: {str(e)}")
        raise HttpError(500, "Internal Server Error")


@router.put("/profiles/{profile_id}/edit", response=WorkProfileSchema, auth=TokenAuth())
def edit_profile(request, profile_id: int, payload: CreateProfileSchema):
    """
    Edits an existing work profile. If the profile is set to active, all other profiles will be deactivated.
    """
    try:
        profile = WorkProfile.objects.get(id=profile_id)

        if payload.is_active:
            WorkProfile.objects.filter(is_active=True).exclude(id=profile_id).update(
                is_active=False
            )

        profile.name = payload.name
        profile.profile_type = payload.profile_type
        profile.is_active = payload.is_active

        # Обновляем URL-адреса
        profile.selected_urls.clear()
        for url in payload.selected_urls:
            profile_url, _ = ProfileURL.objects.get_or_create(url=url)
            profile.selected_urls.add(profile_url)

        # Обновляем приложения
        profile.selected_apps.clear()
        for app_name in payload.selected_apps:
            app, _ = Application.objects.get_or_create(name=app_name)
            profile.selected_apps.add(app)

        profile.save()
        return {
            "id": profile.id,
            "name": profile.name,
            "profile_type": profile.profile_type,
            "is_active": profile.is_active,
            "selected_urls": [url.url for url in profile.selected_urls.all()],
            "selected_apps": [app.name for app in profile.selected_apps.all()],
        }
    except WorkProfile.DoesNotExist:
        raise HttpError(404, "Profile not found")
    except Exception as e:
        logger.error(f"Error editing profile: {str(e)}")
        raise HttpError(500, "Internal Server Error")


@router.put(
    "/profiles/{profile_id}/set-active", response=WorkProfileSchema, auth=TokenAuth()
)
def set_active_profile(request, profile_id: int):
    """
    Sets a specific work profile as active. Deactivates all other profiles.
    """
    try:
        # Проверяем, существует ли профиль с данным profile_id
        profile = WorkProfile.objects.get(id=profile_id)

        # Деактивируем все другие профили
        WorkProfile.objects.filter(is_active=True).update(is_active=False)

        # Устанавливаем текущий профиль как активный
        profile.is_active = True
        profile.save()

        # Возвращаем данные профиля с преобразованными полями для selected_urls и selected_apps
        return {
            "id": profile.id,
            "name": profile.name,
            "profile_type": profile.profile_type,
            "is_active": profile.is_active,
            "selected_urls": [url.url for url in profile.selected_urls.all()],
            "selected_apps": [app.name for app in profile.selected_apps.all()],
        }

    except WorkProfile.DoesNotExist:
        raise HttpError(404, "Profile not found")
    except Exception as e:
        logger.error(f"Error setting active profile: {str(e)}")
        raise HttpError(500, "Internal Server Error")


@router.delete("/profiles/{profile_id}/delete", auth=TokenAuth())
def delete_profile(request, profile_id: int):
    """
    Deletes a work profile. If the deleted profile was active, another profile is activated (if available).
    """
    try:
        profile = WorkProfile.objects.get(id=profile_id)
        was_active = profile.is_active
        profile.delete()

        if was_active:
            other_profile = WorkProfile.objects.filter(is_active=False).first()
            if other_profile:
                other_profile.is_active = True
                other_profile.save()

        return {"message": "Profile deleted successfully"}
    except WorkProfile.DoesNotExist:
        raise HttpError(404, "Profile not found")
    except Exception as e:
        logger.error(f"Error deleting profile: {str(e)}")
        raise HttpError(500, "Internal Server Error")


@router.get("/profiles", response=List[WorkProfileSchema], auth=TokenAuth())
def get_all_profiles(request):
    """
    Retrieves a list of all work profiles.
    """
    profiles = WorkProfile.objects.all()
    return [
        {
            "id": profile.id,
            "name": profile.name,
            "profile_type": profile.profile_type,
            "is_active": profile.is_active,
            "selected_urls": [url.url for url in profile.selected_urls.all()],
            "selected_apps": [app.name for app in profile.selected_apps.all()],
        }
        for profile in profiles
    ]


@router.get("/profiles/{profile_id}", response=WorkProfileSchema, auth=TokenAuth())
def get_profile(request, profile_id: int):
    """
    Retrieves details of a specific work profile by ID.
    """
    try:
        profile = WorkProfile.objects.get(id=profile_id)
        return {
            "id": profile.id,
            "name": profile.name,
            "profile_type": profile.profile_type,
            "is_active": profile.is_active,
            "selected_urls": [url.url for url in profile.selected_urls.all()],
            "selected_apps": [app.name for app in profile.selected_apps.all()],
        }
    except WorkProfile.DoesNotExist:
        raise HttpError(404, "Profile not found")


@router.get("/applications", response=List[ApplicationSchema], auth=TokenAuth())
def get_applications(request):
    """
    Retrieves the list of available applications.
    """
    applications = Application.objects.all()
    return applications


@router.get("/update/check", response=UpdateStatusResponse)
def check_for_updates(request):
    """
    Checks if updates are available via apt.
    """
    try:
        updates_available = update_service.check_updates()
        return updates_available
    except Exception as e:
        logger.error(f"Error checking for updates: {str(e)}")
        raise HttpError(500, {"error": "Failed to check for updates."})


@router.post("/update/start")
def start_update(request):
    success, message = UpdateService.perform_update()
    return {"success": success, "message": message}


@router.get("/update/status")
def get_update_status(request):
    status = UpdateService.get_update_status()
    return status
